#include<iostream>
#include<complex>
#define complex complex<double>

#include<bar.h>
#include<factorial.h>
#include<fluid.h>
#include<grid.h>
#include<hamop.h>
#include<wavefunction.h>
#include<xml_parse.h>
#include<ylm.h>
//
// potentials, nuclear_charge etc. are defined in potentials.cc
//
#include "potentials.cc"

int main(int argc, char **argv)
{
  complex imagi(0.0,1.0);

  char cstr_fname_logfi[1000], cstr_fname_wf_ini[1000], 
       cstr_fname_obser[1000], cstr_fname_wf[1000],
       cstr_fname_vpot_xy[1000], cstr_fname_yield[1000],
       cstr_fname_im_info[1000], cstr_fname_re_info[1000];

  double real_timestep, time, E_tot, delt_x,
         N, yield_N, yield_P,
         x_expect, y_expect;
  complex polplus, polminus;
  int ivecpotflag, iinitmode, isave_wf, i, me=0, iv;
  int *really_propagate;
  long lno_of_ts, ts, ngps_x, ngps_y, ngps_z, dimens, 
       ldumpwidth, lcount_dwidth, lks_norm;
  FILE *file_logfi, *file_wf_ini, *file_obser, *file_wf, 
       *file_vpot_xy, *file_yield, *file_info;

  complex timestep;

  grid g, g_load;
  hamop hamilton;
  wavefunction staticpot, wf, wf_load;
  complex P;

  fprintf(stdout, " CIRC_RE: Real-time propagation for hydrogen atom\n");
  fprintf(stdout, " (C) Copyright by Bauer D and Koval P, Heidelberg (2005)\n");
  fprintf(stdout, " -------------------------------------------------------\n");

    
  //
  // input (result from imaginary propagation)
  //
  sprintf(cstr_fname_wf_ini,"./res/circ_im_wf_fin-44-1000-1.50e-01.dat");
  sprintf(cstr_fname_im_info,"./res/circ_im_info-44-1000-1.50e-01.dat");


  isave_wf      = 1; // 1 -- save or 0 -- don't save wavefunctions
  iv            = 1;          // verbosity of stdout


  //
  // get info from imaginary propagation
  //
  file_info = fopen(cstr_fname_im_info, "r");
  if (file_info==NULL) {
    fprintf(stderr, "%s can't be opened!\n", cstr_fname_im_info); exit(-10);}  

  xml_parse_fgetd(&nuclear_charge,file_info,"<nuclear_charge>",
    "</nuclear_charge>");
  xml_parse_fgetl(&ngps_x, file_info, "<ngps_x>", "</ngps_x>");
  xml_parse_fgetl(&ngps_y, file_info, "<ngps_y>", "</ngps_y>");
  xml_parse_fgetl(&ngps_z, file_info, "<ngps_z>", "</ngps_z>");
  xml_parse_fgetd(&delt_x, file_info, "<delt_x>", "</delt_x>");
  xml_parse_fgetl(&dimens, file_info, "<dimens>", "</dimens>");
  fclose(file_info);
  
  // *** declare the grid for propagation ***
  g.set_dim(dimens);
  g.set_ngps(ngps_x, 4, ngps_z); 
  g.set_delt(delt_x);
  g.set_offs(0,0,0);

  // *** declare the grid for load ***
  g_load.set_dim(dimens);
  g_load.set_ngps(ngps_x,ngps_y,ngps_z); 
  g_load.set_delt(delt_x,0.0,0.0);
  g_load.set_offs(0,0,0);
 
  sprintf(cstr_fname_yield, "./res/circ_re_yield-%lf-%d-%ld-%.2le.dat",w,
    dimens, ngps_x, delt_x);
  sprintf(cstr_fname_logfi, "./res/circ_re-%lf-%d-%ld-%.2le.log",w,
    dimens, ngps_x, delt_x);
  
  // *** create some files with appropriate appendices
  file_logfi      = fopen(cstr_fname_logfi,"w");
  if (file_logfi==NULL)
  {fprintf(stdout, "%s cant be opened!\n", cstr_fname_logfi); exit(-2);}

  if (iv!=0)
  {
     fprintf(stdout, "%s will be (re)written.\n", cstr_fname_yield);
     fprintf(stdout, "%s will be (re)written.\n", cstr_fname_logfi);
  }

  file_yield    = fopen(cstr_fname_yield,"w");
  if (file_yield==NULL)
  {fprintf(stdout, "%s cant be opened!\n", cstr_fname_yield); exit(-5);}

  file_logfi      = fopen(cstr_fname_logfi,"w");
  if (file_logfi==NULL)
  {fprintf(stdout, "%s cant be opened!\n", cstr_fname_logfi); exit(-2);}
 

  
  // electric field strength (used in potentials.cc)
  E_0     = 3.774e-3;


  // 
  // variables
  // 
  real_timestep = 0.05;
  lno_of_ts     = 2*long(81.0/(16.0*E_0*real_timestep));
  ldumpwidth    = 20;


  lks_norm      = (long) (20.0 / delt_x);
  if (lks_norm>ngps_x) lks_norm=ngps_x;

  sprintf(cstr_fname_obser, "./res/circ_re_obser-%lf-%d-%ld-%.2le.dat",
    w,g.dimens(),g.ngps_x(),delt_x);
  if (isave_wf==1)
  {
    sprintf(cstr_fname_wf,"./res/circ_re_wf-%lf-%d-%ld-%.2le.dat",
    w,g.dimens(),g.ngps_x(),delt_x);

    sprintf(cstr_fname_re_info,"./res/circ_re_info-%lf-%d-%ld-%.2le.dat",
    w,g.dimens(),g.ngps_x(),delt_x);
  }
  sprintf(cstr_fname_vpot_xy,"./res/circ_re_vpot_xy-%lf-%d-%.2le.dat",
    w,g.dimens(),delt_x);


  // *** declare rest of variables ***
  hamilton.init(g,vecpot_x,vecpot_y,vecpot_z,
                  scalarpotx,scalarpoty,scalarpotz,
                  imagpot,field);

  // this is the linear and constant part of the Hamiltonian
  staticpot.init(g.size()); 
  staticpot.calculate_staticpot(g,hamilton);

  // *** wavefunction array 
  wf.init(g.size()); 
  wf_load.init(g_load.size());

  // *** create some files with appropriate appendices
  file_obser = fopen(cstr_fname_obser,"w");
  if (file_obser==NULL)
  {fprintf(stderr, "err: %s cant be opened!\n", cstr_fname_obser); exit(-3);}

  file_wf_ini    = fopen(cstr_fname_wf_ini,"r");
  if (file_wf_ini==NULL)
  {fprintf(stderr, "err: %s cant be opened!\n", cstr_fname_wf_ini); exit(-4);}


  if (iv!=0)
  {
     fprintf(stdout, "%s will be read.\n", cstr_fname_wf_ini);
     fprintf(stdout, "%s will be (re)written.\n", cstr_fname_obser);
     if (isave_wf==1)
     {
       fprintf(stdout, "%s will be (re)written.\n", cstr_fname_wf);
       fprintf(stdout, "%s will be (re)written.\n", cstr_fname_re_info);
     }
     fprintf(stdout, "%s will be (re)written.\n", cstr_fname_vpot_xy);
  }

  ts=0;
  
  // *** wavefunction initialization ***
  wf_load.init(g_load,file_wf_ini,0,iv);
  wf.regrid(g, g_load, wf_load);    
  fclose(file_wf_ini);

  if (isave_wf==1)
  {
    file_wf    = fopen(cstr_fname_wf,"w");
    if (file_wf==NULL)
    {fprintf(stderr, "err: %s cant be opened!\n", cstr_fname_wf); exit(-8);}
    wf.dump_to_file_sh(g,file_wf,1); // initial wf is saved
  }
  fprintf(stdout,"Norm: %le\n",wf.norm(g));

  
  //
  // write to log file
  //
  fprintf(file_logfi,"Real-time propagation\n");
  fprintf(file_logfi,"Grid: \n", g.ngps_x());
  fprintf(file_logfi,"g.ngps_x() = %ld\n", g.ngps_x());
  fprintf(file_logfi,"g.ngps_y() = %ld\n", g.ngps_y());
  fprintf(file_logfi,"g.ngps_z() = %ld\n", g.ngps_z());
  fprintf(file_logfi,"g.dimens() = %ld\n\n", g.dimens());
  fprintf(file_logfi,"g.delt_x() = %20.15le\n", g.delt_x());

  fprintf(file_logfi,"real_timestep     = %20.15le\n", real_timestep);
  fprintf(file_logfi,"lno_of_ts         = %ld\n", lno_of_ts);
  fprintf(file_logfi,"nuclear_charge    = %20.15le\n", nuclear_charge);
  fprintf(file_logfi,"cstr_fname_wf_ini = %s\n", cstr_fname_wf_ini);
  fprintf(file_logfi,"cstr_fname_obser  = %s\n", cstr_fname_obser);
  if (isave_wf==1) 
    fprintf(file_logfi,"cstr_fname_wf = %s\n", cstr_fname_wf);
  fprintf(file_logfi,"n_c        = %lf\n", n_c);  
  fprintf(file_logfi,"w        = %10.5le\n", w);
  fprintf(file_logfi,"E_0      = %10.5le\n", E_0);

  fflush(file_logfi);

  fprintf(stdout,"ngpsx*deltx  = %10.5le\n", g.ngps_x()*g.delt_x());
  fprintf(stdout,"lno_of_ts    = %ld\n", lno_of_ts);

  file_vpot_xy = fopen(cstr_fname_vpot_xy, "w");
  if (file_vpot_xy==NULL)
  {fprintf(stdout, "%s cant be opened!\n", cstr_fname_vpot_xy); exit(-7);}

  time = 0.0;
  lcount_dwidth = 0;


  // write vector potential to file
  for (ts=0; ts<lno_of_ts; ts++)
  {
     time = time + real_timestep;
     
     if (lcount_dwidth==0)
     fprintf(file_vpot_xy, "%10.5le %10.5le %10.5le\n", 
       time, vecpot_x(time, me), vecpot_y(time, me));

     lcount_dwidth = lcount_dwidth + 1;
     if (lcount_dwidth>ldumpwidth) lcount_dwidth=0;
  }
  fclose(file_vpot_xy);


  // ********************************************************
  // ***** real time propagation ****************************
  // ********************************************************


  E_tot = 0.0;
  timestep=complex(real_timestep, 0.0);
  time = 0.0;
  lcount_dwidth = 0;


  for (ts=0; ts<lno_of_ts; ts++){
    time = time + real_timestep;
    
    //
    // observables and output
    //
    if (lcount_dwidth==0)
    {
    //
    // calculate total energy, projection onto initial state,
    // norm, <x>, and <y>
    // 
    E_tot = real(wf.energy(0.0,g,hamilton,me,staticpot,nuclear_charge));
    P     = wf.project(g, g_load, wf_load, 0);
    N     = wf.norm(g);
    polplus=wf.expect_cycl_pol_plus(g);
    polminus=wf.expect_cycl_pol_minus(g);
    x_expect=real(0.5*(polplus+polminus));
    y_expect=real(-0.5*imagi*(polplus-polminus));

    fprintf(file_obser,"%10.5le %20.15le %20.15le %20.15le %20.15le %20.15le\n",time,E_tot,real(conj(P)*P),N,x_expect,y_expect);
    }

    //
    // stdout
    // 
    if ((iv==1) && (lcount_dwidth==0) && is_time(0.25))
      fprintf(stdout,"%10.5le %10.5le %10.5le %10.5le %ld\n",time,E_tot,
        real(conj(P)*P), N, ts);

    //
    // propagate one step forward in (real) time.
    // 
    wf.propagate(timestep, time, g, hamilton, me, staticpot,
		 0, nuclear_charge);

    //
    // decide whether to do the output during the next step.
    // 
    lcount_dwidth = lcount_dwidth + 1;
    if (lcount_dwidth>ldumpwidth) lcount_dwidth=0;

  } // end of real-time-propagation loop

  fclose(file_obser);

  yield_N = (1.0 - N);
  yield_P = (1.0 - real(conj(P)*P));

  fprintf(file_yield, "%10.5le %20.15le %20.15le\n",
    E_0, yield_N, yield_P);

  free(really_propagate);

  if (isave_wf==1) 
  {
    wf.dump_to_file_sh(g,file_wf,1); // final wf is saved
    fclose(file_wf); 
 
    file_info = fopen(cstr_fname_re_info, "w");
    fprintf(file_info,"<nuclear_charge>%10.5le</nuclear_charge>\n",
      nuclear_charge);
    fprintf(file_info,"<ngps_x>%ld</ngps_x>\n",g.ngps_x());
    fprintf(file_info,"<ngps_y>%ld</ngps_y>\n",g.ngps_y());
    fprintf(file_info,"<ngps_z>%ld</ngps_z>\n",g.ngps_z());
    fprintf(file_info,"<delt_x>%10.5le</delt_x>\n",g.delt_x());
    fprintf(file_info,"<dimens>%ld</dimens>\n",g.dimens());
    fprintf(file_info,"<w>%10.5le</w>\n", w);
    fclose(file_info);
  }

  if (iv!=0)
  {
    fprintf(stdout, "%s was read.\n",   cstr_fname_wf_ini);
    fprintf(stdout, "%s is written.\n", cstr_fname_obser);
    if (isave_wf==1) 
    {
      fprintf(stdout, "%s is written.\n", cstr_fname_wf);
      fprintf(stdout, "%s will be (re)written.\n", cstr_fname_re_info);
    }
    fprintf(stdout, "%s is written.\n", cstr_fname_vpot_xy);
  }


  fclose(file_logfi);
  fclose(file_yield);

  if (iv!=0)
  {  fprintf(stdout, "%s is written.\n", cstr_fname_logfi);
     fprintf(stdout, "%s is written.\n", cstr_fname_yield);
  }

  fprintf(stdout, "Hasta la vista...\n");
}

//
// end of main program
//
