#include<iostream>
#include<complex>
#define complex complex<double>

#include<bar.h>
#include<fluid.h>
#include<grid.h>
#include<hamop.h>
#include<wavefunction.h>
//
// functions determining the potentials,
// nuclear_charge are defined in the potentials.cc
//
#include "potentials.cc"

int main(int argc, char **argv)
{
  //
  // variables
  //
  char cstr_fname_logfi[1000], cstr_fname_wf_ini[1000], 
       cstr_fname_obser[1000], cstr_fname_wf[1000], cstr_fname_wf_read[1000],
       cstr_fname_info[1000],  cstr_fname_hpot0[1000], cstr_fname_xpot0[1000];

  double real_timestep, time, E_tot, E_sp;
  int iinitmode, i, me, iv, islateronly, ich;
  long lno_of_ts, ts;
  FILE *file_logfi, *file_wf_ini, *file_obser_real,
       *file_wf, *file_info, *file;

  complex timestep, x_energy, hartree_energy;

  fluid degener, ms, ells, V_ee_0, Lambdavector, U_H0, U_x0;

  grid g, g_read;
  hamop hamilton;
  wavefunction staticpot, V_ee_1, V_ee_2, wf, wf_read, orb_energs, 
    orb_hartreesandexchange;
  
  fprintf(stdout, "CLUSTER_RE: Imaginary-time propagation for Na cluster\n");
  fprintf(stdout, "(C) Copyright by Bauer D and Koval P, Heidelberg (2005)\n");
  fprintf(stdout, "-------------------------------------------------------\n");

  //
  // input
  //
  // *** declare the grid for reading the initial wavefunction
  g_read.set_dim(34); // only 34 (linear polariz.) works in Kohn-Sham mode
  g_read.set_ngps(8000,2,3); // choose 2 angular momenta and 3 orbitals
  g_read.set_delt(0.2);
  g_read.set_offs(0,0,0);

  // *** declare the grid for propagation ***
  nuclear_charge = 0.0; // in the cluster case no correction at r=0 arises!
  g.set_dim(34); // only 34 (linear polariz.) works in Kohn-Sham mode
  g.set_ngps(8000,4,3); // choose 2 angular momenta and 3 orbitals
  g.set_delt(0.2);
  g.set_offs(0,0,0);

  //
  // calculation of the KLI potential
  //
  islateronly  = 0;


  //
  // amplitude of the imaginary potential
  //
  imagi_ampl=100.0;


  // verbosity of stdout
  iv           = 1; 

  sprintf(cstr_fname_logfi, "./res/Na8_re-%3.3lf.log", g.delt_x());
  sprintf(cstr_fname_obser, "./res/Na8_re_obser-%3.3lf.dat", g.delt_x());
  sprintf(cstr_fname_wf,"./res/Na8_re_wf-%3.3lf.dat", g.delt_x());
  sprintf(cstr_fname_info,  "./res/Na8_re_info-%3.3lf.dat", g.delt_x());
  sprintf(cstr_fname_hpot0, "./res/Na8_re_H_0-%3.3lf.dat", g.delt_x());
  sprintf(cstr_fname_xpot0, "./res/Na8_re_X_0-%3.3lf.dat", g.delt_x());

  sprintf(cstr_fname_wf_ini,"./res/Na8_im_wf_fin-0.200.dat");

  //
  // executable part
  // 

  // Number of real time steps
  lno_of_ts = 254; 
  degener.init(g.ngps_z());
  ms.init(g.ngps_z());
  ells.init(g.ngps_z());
  int really_propagate[g.ngps_z()];


  really_propagate[0] = 1; // orbital 0 is to be propagated (not frozen!)
  really_propagate[1] = 1; // orbital 1 is to be propagated (not frozen!)
  really_propagate[2] = 1; // orbital 2 is to be propagated (not frozen!)

  degener[0]          = 2; // two electrons in 1s shell
  degener[1]          = 4; // two electrons in 2s shell
  degener[2]          = 2; // six electrons in 2p subshell

  ms[0]               = 0; // m quantum number of orbital 0 
  ms[1]               = 1; // m quantum number of orbital 1 
  ms[2]               = 0; // whether one puts 0,1, or -1 here 
                           // does not matter for the ground state 

  ells[0]             = 0; // l quantum number of orbital 0 
  ells[1]             = 1; // l quantum number of orbital 1 
  ells[2]             = 1; // l quantum number of orbital 2 

  me                  = 0; // dummy


  // *** declare rest of variables ***
  real_timestep=0.05 ;


  // *** the laser field strength 
  E_0=1.0;

  // this is the linear, constant part of the Hamiltonian
  staticpot.init(g.size()); 
  hamilton.init(g,vecpot_x,vecpot_y,vecpot_z,scalarpotx,scalarpoty,scalarpotz,
    imagpot,field);

  staticpot.calculate_staticpot(g,hamilton);



  // *** the wavefunction array for reading 
  wf_read.init(g_read.size()); 

  // *** the wavefunction array for propagation
  wf.init(g.size()); 

  Lambdavector.init(g.ngps_x()); 
  U_H0.init(g.ngps_x());
  U_x0.init(g.ngps_x());
  V_ee_0.init(g.ngps_x());
  V_ee_1.init(g.ngps_x());
  V_ee_2.init(g.ngps_x());

  orb_energs.init(g.ngps_z());
  orb_hartreesandexchange.init(g.ngps_z());

  // *** create some files with appropriate appendices
  file_logfi      = fopen(cstr_fname_logfi,"w");
  if (file_logfi==NULL)
  {fprintf(stderr, "%s cant be opened!\n", cstr_fname_logfi); exit(-2);}

  file_obser_real = fopen(cstr_fname_obser,"w");
  if (file_obser_real==NULL)
  {fprintf(stderr, "%s cant be opened!\n", cstr_fname_obser); exit(-3);}

  file_wf_ini    = fopen(cstr_fname_wf_ini,"r");
  if (file_wf_ini==NULL)
  {fprintf(stderr, "%s cant be opened!\n", cstr_fname_wf_ini); exit(-4);}

  if (iv!=0)
  {
     fprintf(stderr, "%s will be (re)written.\n", cstr_fname_logfi);
     fprintf(stderr, "%s will be opened.\n", cstr_fname_wf_ini);
     fprintf(stderr, "%s will be (re)written.\n", cstr_fname_obser);
     fprintf(stderr, "%s will be (re)written.\n", cstr_fname_wf);
     fprintf(stderr, "%s will be (re)written.\n", cstr_fname_info);
  }

  ts=0;
  
  // *** initialization ***
  wf_read.init(g_read,file_wf_ini,0,iv); // read wavefunction from file
  wf.regrid(g, g_read, wf_read); // regrid it to the bigger grid for
                                 // real time propagation with laser field 
  fclose(file_wf_ini);


  fprintf(stdout,"Norm of KS-orbital: %le\n",wf.total_ks_norm(g,degener));
  
  //
  // Fill the log file
  //
  fprintf(file_logfi,"Real-time propagation\n");
  fprintf(file_logfi,"Grid: \n", g.ngps_x());
  fprintf(file_logfi,"g.ngps_x() = %ld\n", g.ngps_x());
  fprintf(file_logfi,"g.ngps_y() = %ld\n", g.ngps_y());
  fprintf(file_logfi,"g.ngps_z() = %ld\n", g.ngps_z());
  fprintf(file_logfi,"g.dimens() = %ld\n\n", g.dimens());
  fprintf(file_logfi,"g.delt_x() = %15.10le\n", g.delt_x());

  fprintf(file_logfi,"real_timestep     = %15.10le\n", real_timestep);
  fprintf(file_logfi,"lno_of_ts         = %ld\n", lno_of_ts);
  fprintf(file_logfi,"nuclear_charge    = %15.10le\n", nuclear_charge);
  fprintf(file_logfi,"iinitmode         = %d\n", iinitmode);
  fprintf(file_logfi,"cstr_fname_wf_ini = %s\n", cstr_fname_wf_ini);
  fprintf(file_logfi,"cstr_fname_obser  = %s\n", cstr_fname_obser);
  fprintf(file_logfi,"cstr_fname_wf = %s\n", cstr_fname_wf);
  fprintf(file_logfi,"cstr_fname_info   = %s\n", cstr_fname_info);
  fflush(file_logfi);

  // store initial wavefunction on propagation grid
  file_wf    = fopen(cstr_fname_wf,"w");
  if (file_wf==NULL)
  {fprintf(stderr, "%s cant be opened!\n", cstr_fname_wf); exit(-5);}
  wf.dump_to_file_sh(g,file_wf,1);


  // ********************************************************
  // ***** Real-time propagation ****************************
  // ********************************************************

  E_tot = 0.0;
  timestep=complex(real_timestep,0.0);  
  time = 0.0;

  for (ts=0; ts<lno_of_ts; ts++)
  {
    time =  time + real_timestep;

    //
    // computation of monopole term for Hartree and exchange potential
    //
      Lambdavector = wf.calculate_Lambda(g,degener);
      U_H0 = wf.calculate_hartree_zero(g,Lambdavector);
      U_x0 = wf.calculate_kli_zero(g,Lambdavector,ells,ms,
        degener,islateronly,0);
      V_ee_0 = U_H0 + U_x0;


    wf.propagate(timestep, time, g, hamilton, me, staticpot,
		 V_ee_0, V_ee_1, V_ee_2, ms, nuclear_charge, really_propagate);
    

    //
    // calculate energies
    // 
    orb_energs = wf.orbital_energies(0.0, g, hamilton,me,
			      staticpot, nuclear_charge);
    orb_hartreesandexchange = wf.orbital_hartrees(0.0,g,me,U_H0+U_x0);
    E_sp = wf.totalenergy_single_part(g,orb_energs, degener);
    hartree_energy = wf.totalenergy_hartree(g,Lambdavector,U_H0);
    x_energy=wf.totalenergy_exact_x(g,ells,degener);
    E_tot = real(E_sp + hartree_energy + x_energy);
    //
    // output to observables file
    //    
    fprintf(file_obser_real,"%i %15.10le %15.10le %15.10le %15.10le ",
	    ts,E_tot,E_sp,real(hartree_energy),real(x_energy));
    for (i=0;i<g.ngps_z();i++)  // loop over KS orbitals
      fprintf(file_obser_real,"%15.10le ",real(orb_energs[i]+orb_hartreesandexchange[i]));
    fprintf(file_obser_real,"\n");

    //
    // output to stdout
    //
    fprintf(stdout,"time step: %i\n total energy: %15.10le\n single-particle:  %15.10le\n Hartree: %15.10le\n exchange: %15.10le\n ",
	    ts,E_tot,E_sp,real(hartree_energy),real(x_energy));
    for (i=0;i<g.ngps_z();i++) 
      fprintf(stdout,"energy of orbital %i: %15.10le\n ",i,real(orb_energs[i]+orb_hartreesandexchange[i]));
    fprintf(stdout,"\n");
    
  }
 
  
  fclose(file_obser_real);

  file    = fopen(cstr_fname_hpot0, "w");
  if (file==NULL){
    fprintf(stderr, "%s cant be opened!\n", cstr_fname_hpot0); exit(-6);}
  U_H0.dump_to_file_sh(g,file,1);
  fclose(file);

  file    = fopen(cstr_fname_xpot0, "w");
  if (file==NULL){
    fprintf(stderr, "%s cant be opened!\n", cstr_fname_xpot0); exit(-6);}
  U_x0.dump_to_file_sh(g,file,1);
  fclose(file);

  // store final wavefunction
  wf.dump_to_file_sh(g,file_wf,1);
  fclose(file_wf);

  file_info = fopen(cstr_fname_info, "w");
  fprintf(file_info,"<nuclear_charge>%15.10le</nuclear_charge>\n",
    nuclear_charge);
  fprintf(file_info,"<ngps_x>%ld</ngps_x>\n",g.ngps_x());
  fprintf(file_info,"<ngps_y>%ld</ngps_y>\n",g.ngps_y());
  fprintf(file_info,"<ngps_z>%ld</ngps_z>\n",g.ngps_z());
  fprintf(file_info,"<delt_x>%15.10le</delt_x>\n",g.delt_x());
  fprintf(file_info,"<dimens>%ld</dimens>\n",g.dimens());
  fprintf(file_info,"<islateronly>%d</islateronly>\n",islateronly);
  fclose(file_info);

  if (iv!=0)
  {
     fprintf(stderr, "%s is written.\n", cstr_fname_logfi);
     fprintf(stderr, "%s is written.\n", cstr_fname_wf_ini);
     fprintf(stderr, "%s is written.\n", cstr_fname_obser);
     fprintf(stderr, "%s is written.\n", cstr_fname_wf);
     fprintf(stderr, "%s is written.\n", cstr_fname_info);
     fprintf(stderr, "%s is written.\n", cstr_fname_hpot0);
     fprintf(stderr, "%s is written.\n", cstr_fname_xpot0);
  }

  fprintf(stdout, "Hasta la vista...\n");
  fclose(file_logfi);
}

//
// end of main program
//
